#include "alaw_ulaw.h"

static unsigned char linear2law(short sample, char law);
static int ulaw2linear(unsigned char sample);
static int alaw2linear(unsigned char sample);

static unsigned char linear2law(short sample, char law)
{
    int best=-1;
    int i = 0;
    int diff = 0;
    int best_diff = 0;
    
    while(i<256)
    {
        // law='u' -> u-law
        // law='a' -> a-law
        diff = (law == 'u') ? isdn_audio_ulaw_to_s16[i] : isdn_audio_alaw_to_s16[i] - sample;

        if (diff<0)
            diff = 0 - diff;

        if (diff < best_diff || best < 0)
        {
            best_diff = diff;
            best = i;
        }

        i++;
    }

    return(best);
}

static int ulaw2linear(unsigned char sample)
{
    short t;

    sample = ~sample;

    t = ((sample & 0xf) << 3) + 0x84;
    t <<= ((unsigned)sample & 0x70) >> 4;

    return ((sample & 0x80) ? (0x84 - t) : (t - 0x84));
}

static int alaw2linear(unsigned char sample) 
{
    signed short r = 0, sign = 1;

    // Must reverse bit order.
    sample = ((sample & 0xaaaa) >> 1) | ((sample & 0x5555) << 1);
    sample = ((sample & 0xcccc) >> 2) | ((sample & 0x3333) << 2);
    sample = (sample >> 4) | (sample << 4);
    if (sample & 0x80) {
        sample ^= 0xd5;            // Flip bits for positive values
    } else {
        sample ^= 0x55;            // Flip bits for negative values
        sign = -1;            // Remember sign
    }
    switch (sample & 0x70) {
    case 0x70:                // Segment 7 (0x800..0xfff)
        r = 0x840 | ((sample & 0xf) << 7);
        break;
    case 0x60:                // Segment 6 (0x400..0x7ff)
        r = 0x420 | ((sample & 0xf) << 6);
        break;
    case 0x50:                // Segment 5 (0x200..0x3ff)
        r = 0x210 | ((sample & 0xf) << 5);
        break;
    case 0x40:                // Segment 4 (0x100..0x1ff)
        r = 0x108 | ((sample & 0xf) << 4);
        break;
    case 0x30:                // Segment 3 (0x080..0x0ff)
        r = 0x084 | ((sample & 0xf) << 3);
        break;
    case 0x20:                // Segment 2 (0x040..0x07f)
        r = 0x042 | ((sample & 0xf) << 2);
        break;
    default:                  // Segment 1 (0x000..0x03f)
        r = 0x001 | ((sample & 0x1f) << 1);
    }
    return (r*sign) << 3;
}

// Build table .. makes encoding pretty fast :)
void create_table_buffer(int use_ulaw)
{   
    signed char *_linear16_2_law = &linear16_2_law[32768];
    char law[2];    
    long i;
    
    if(use_ulaw)
        law[0]='u';
    else
        law[0]='a';

    law[1] = '\0';

    // build law->linear16
    for(i = 0; i < 65535; i++)                    
        _linear16_2_law[i-32768] = linear2law((short)i-32768, law[0]);    

    // build linear16->law    
    for(i = 0; i < 256; i++)
    {
        if(use_ulaw)
            law_2_linear16[i] = ulaw2linear((unsigned char)i);
        else
            law_2_linear16[i] = alaw2linear((unsigned char)i);
    }
}
